<?php /* seeder.php
This file is part of BlogTorrent http://www.blogtorrent.com/
nassar 'at' downhillbattle 'dot' org
Licensed under the terms of the GNU GPL */
require_once("include.php");

/*

We use a custom frontend for the official BitTorrent client to perform
server-side seeding. It forks itself off as a background process and
writes stats to a file automatically, so we don't need to depend on
nohup or something similar being around.

We chose to use a Python based BT client, so that platforms we don't
explictly support can

A stripped down Python interpreter for Linux x86 is included in the
Blog Torrent distribution, so server sharing should work automatically
on most Linux servers. If you have a Mac OS X or UNIX server, you'll
need to tell Blog Torrent where it can find Python.

Usage: 

setup() checks to see if the seeder is configured and setup the data
        directories, etc.
setupHelpMessage() returns an error message explaining the last
                   failure of setup()
spawn($torrentfile) starts seeding a file
pause($torrentfile) stops seeding a file
stop($torrentfile) stops seeding a file and deletes the cached copy

*/

class ServerSideSeeder {

  function ServerSideSeeder() {
    $this->enabled = false;
    $this->problem = "";
    $this->supported_platforms = array("linux-386");
  }
  //Private function  
  //Unpacks a zipfile to the $prefix directory
  function unzipFile($file,$prefix="data") {
    $file = dirname($_SERVER['SCRIPT_FILENAME']).'/'.$file;
    $zip = zip_open($file);
    if ($zip) {
      while ($entry = zip_read($zip)) {
	$name = zip_entry_name($entry);
	$size = zip_entry_filesize($entry);
	$dir = dirname($name);
	$full_curdir = '';
	if ($dir != '.') 
	  foreach (explode('/',$dir) as $curdir) {
	    if (!strlen($full_curdir))
	      $full_curdir = $curdir;
	    else
	      $full_curdir .= '/'.$curdir;
	    if (!file_exists($prefix.'/'.$full_curdir))
	      mkdir($prefix.'/'.$full_curdir);
	  }
	if (substr($name,strlen($name)-1,1) == '/') {
	  mkdir($prefix.'/'.$name);
	}
	else
	  if (zip_entry_open($zip,$entry,"rb")) {
	    $fp = fopen($prefix.'/'.$name,"wb");
	    fwrite($fp,zip_entry_read($entry,$size),$size);
	    fclose($fp);
	    zip_entry_close($entry);
	  }
      }
      zip_close($zip);
      return true;
    } else {
      return false; //This version of PHP doesn't support zip
    }
  }

  //Returns true if server side seeding is enabled
  function enabled() {
    return $this->enabled;
  }

  //Private Function
  function unzipBT() {
    return $this->unzipFile("btseeder.zip");
  }

  //Private function
  function unzipPython() {
    return $this->unzipFile("python.zip");
  }

  function findPython() {
    global $settings;
    if (strlen($settings["sharing_actual_python"]))
      return $settings["sharing_actual_python"];

    if (isset($settings["sharing_python"]) &&
	strlen($settings["sharing_python"])) {
      $version = exec($settings["sharing_python"]." -V 2>&1");
      if (preg_match('/^Python/',$version)) {
	$settings["sharing_actual_python"] = $settings["sharing_python"];
	save_settings($settings);
	return $settings["sharing_python"];
      }
    }
      

    $prefix = dirname($_SERVER['SCRIPT_FILENAME']);
    foreach ($this->supported_platforms as $platform) {
      $version = exec($prefix."/data/python/$platform/python -V 2>&1");
      if (preg_match('/^Python/',$version)) {
	$settings["sharing_actual_python"] = "PYTHONHOME=$prefix/data/python/$platform PYTHONPATH=$prefix/data/python/$platform $prefix/data/python/$platform/python";
	save_settings($settings);	
	return $settings["sharing_actual_python"];
      }
    }

    $version = exec("python -V 2>&1");
    if (preg_match('/^Python/',$version)) {
      $settings["sharing_actual_python"] = "python";
      save_settings($settings);
      return "python";
    }
    else
      return null;
  }

  function setupHelpMessage() {
    if ($this->problem == "mkdir") {
      $output = "<p>Blog Torrent doesn't have permission to make the server sharing directory automatically. You need to make a writable directory named \"files\" in the data directory. On a typical web server, you might do this by connecting with your ftp client and typing the following commands:</p>\n";
      $output .= "<pre>\n";
      $output .= "cd ".preg_replace('|^(.*[\\/]).*$|','\\1',$_SERVER['SCRIPT_FILENAME'])."data\n";
      $output .= <<<EOD
mkdir files
chmod 777 files
</pre>
<p>Note that giving the directories "777" permissions will allow
anyone on the server to full access those directories. If you share a
server with others, they may be able to tamper with you Blog Torrent
data files if you use these settings. There may be other settings more
appropriate for your server. <b>Please, contact your system administrator
if you have any questions about permissions.</b></p>
EOD;
    } elseif ($this->problem == "unzip") {
      $output = <<<EOD
<p>Blog Torrent cannot unzip the server BitTorrent client. Usually, this is because you are using a version of PHP without zip file support included. You need to unzip "btseeder.zip" and "python.zip" then upload them to the data directory. On a typical web server, you might do this by connecting with your ftp client from the directory with the unzipped files and typing the following commands:</p>
<pre>
EOD;
      $output .= "cd ".preg_replace('|^(.*[\\/]).*$|','\\1',$_SERVER['SCRIPT_FILENAME'])."data\n";
      $output .= <<<EOD
prompt n
mput bt
mput python

EOD;
      foreach ($this->supported_platforms as $platform) {
	$output .= "chmod 755 python/$platform/python\n";
      }
      $output .= "</pre>\n";
    } elseif ($this->problem == "permissions") {
      $output = <<<EOD
<p>Blog Torrent does not have permission to execute the included Python interpreters. You need to change the permissions on the files. On a typical web server, you might do this by connecting with your ftp client and typing the following commands:</p>
<pre>
EOD;
      $output .= "cd ".preg_replace('|^(.*[\\/]).*$|','\\1',$_SERVER['SCRIPT_FILENAME'])."data\n";
      foreach ($this->supported_platforms as $platform) {
	$output .= "chmod 755 python/$platform/python\n";
      }
      $output .= "</pre>\n";
    } elseif ($this->problem == "python") {
      $output = <<<EOD
<p>Blog Torrent could not find a working Python interpreter. Please, specify the location of the python interpreter on your server below. Typical locations are "/usr/bin/python" and "/usr/local/bin/python". Please, contact your system administrator if you need help.</p>
EOD;
    } else {
      $output = "";
    }
    return $output;
  }

  function setup() {
    global $settings;
    $this->enabled = false;

    if (!$settings["sharing_enable"]) {
      $this->problem = "disabled";
      return false;
    }
    error_reporting(0);
    if ((!file_exists('data/files') && !mkdir('data/files',0700)) ||
	(!is_writable('data/files'))) {
      $this->problem = "mkdir";
      return false;
    }
    if ((!file_exists('data/bt') && !$this->unzipBT()) ||
	 (!file_exists('data/python') && !$this->unzipPython())) {
      $this->problem = "unzip";
      return false;
    }
    foreach ($this->supported_platforms as $platform) {
      if ((!is_executable("data/python/$platform/python")) && 
	  (!chmod("data/python/$platform/python",0755))) {
	$this->problem = "permissions";
	return false;
      }
    }
    if (!file_exists('data/files/.htaccess')) {
      $file = fopen('data/files/.htaccess','wb');
      fwrite($file,"deny from all\n");
      fclose($file);
    }

    $this->python = $this->findPython();

    //Production servers should probably comment this out
    error_reporting(E_ALL);

    if ($this->python != null) {
      $this->problem = "";
      $this->enabled = true;
      return true;
    } else {
      $this->problem = "python";
      return false;
    }
  }

  function getSpawnStatus($hash,$update_interval = 15) {
    if (!file_exists("./data/$hash.status"))
      return array();

    $data = array();
    $fp = fopen("./data/$hash.status","r");
    flock($fp,LOCK_EX);
    while ($line = fgets($fp)) {
      $key = preg_replace("/^\s*(.*?)\s*?:.*/s",'\\1',$line);
      $value = preg_replace("/^.*:\s*(.*)\s*$/s",'\\1',$line);
      $data[$key] = $value;
    }
    flock($fp,LOCK_UN);
    fclose($fp);
    $stats = stat("./data/$hash.status");
    $data["Last Update"] = $stats[9]; //Last modification time
    $data["Running"] = (time()-$stats[9] <= $update_interval);
    return $data;
  }

  //Private
  //Does not do locking
  //Does not remove files
  function stop_by_pid($pid) {
    $pid = (integer)$pid;
    if ($pid>0) {
      posix_kill($pid,2); //SIGINT
      sleep(10);
      posix_kill($pid,9); //SIGKILL
    }
    
  }

  //Private
  //Does not do locking
  //Does not remove files
  function stop_by_pid_array($pids) {
    foreach ($pids as $pid) {
      $pid = (integer)$pid;
      if ($pid>0) {
	posix_kill($pid,2); //SIGINT
      }
    }
    sleep(10);
    foreach ($pids as $pid) {
      $pid = (integer)$pid;
      if ($pid>0) {
	posix_kill($pid,9); //SIGKILL
      }
    }
  }

  function stop_array($torrents) {
    global $store;
    $pids = array();
    $delete = array();

    $fp = fopen("./data/spawnlock","wb");
    flock($fp,LOCK_EX);
    
    foreach ($torrents as $torrentfile) {
      $torrentfile = "./torrents/$torrentfile";
      $torrent = bdecode(file_get_contents($torrentfile));
      if (is_array($torrent)) {
	$stats = $this->getSpawnStatus($torrent["sha1"]);
	if (count($stats)>0) {
	  $pids[] = $stats["process id"];
	}
	if (!preg_match('/(\/|^\.\.)/',$torrent["info"]["name"]))
	  $delete[] = "./data/files/".$torrent["info"]["name"];
	$delete[] = "./data/".$torrent["sha1"].".status";
      }
    }

    $this->stop_by_pid_array($pids);
    foreach ($delete as $delme) {
      $this->recursive_remove($delme);
    }

    flock($fp,LOCK_UN);
    fclose($fp);        
  }

  //Recursively removes a directory and all of the files in it
  function recursive_remove($dir) {
    if (file_exists($dir)) {
      if (is_file($dir)) {
	unlink($dir);
      } else {
	$handle = opendir($dir);
	while (false != ($file = readdir($handle))) {
	  if (($file != '.') && ($file != '..'))
	    $this->recursive_remove($dir.'/'.$file);
	}
	closedir($handle);
	rmdir($dir);
      }
    }
  }

  function stop_seeding() {
    global $store;
    $torrents = array();
    $files = $store->getTorrentList();
    foreach ($files as $torrentfile) {
      if (($torrentfile != '.') && ($torrentfile != '..') &&
	  ($torrentfile != '.htaccess')) {
	$torrents[] = $torrentfile;
      }
    }
    $this->stop_array($torrents);

    //Erase any seeded files we might have missed
    $handle = opendir('./data/files/');
    while (false !== ($file = readdir($handle)))
      {
	if (($file != '.') && ($file != '..') &&
	    ($file != '.htaccess')) {
	  $this->recursive_remove("./data/files/$file");
	}
      }
    closedir($handle);
  }

  function pause($torrentfile) {
    $fp = fopen("./data/spawnlock","wb");
    flock($fp,LOCK_EX);    
    $torrentfile = "./torrents/$torrentfile";
    $torrent = bdecode(file_get_contents($torrentfile));
    if (!is_array($torrent)) {
      flock($fp,LOCK_UN);
      fclose($fp);
      return false;
    }
    $stats = $this->getSpawnStatus($torrent["sha1"]);
    $this->stop_by_pid($stats["process id"]);
    flock($fp,LOCK_UN);
    fclose($fp);
  }

  function stop($torrentfile) {
    global $store;
    $fp = fopen("./data/spawnlock","wb");
    flock($fp,LOCK_EX);    
    $torrentfile = "./torrents/$torrentfile";
    $torrent = bdecode(file_get_contents($torrentfile));
    if (!is_array($torrent)) {
      flock($fp,LOCK_UN);
      fclose($fp);
      return false;
    }
    $stats = $this->getSpawnStatus($torrent["sha1"]);
    if (isset($stats["process id"]))
      $this->stop_by_pid($stats["process id"]);

    if (!preg_match('/(\/|^\.\.)/',$torrent["info"]["name"]))
      $this->recursive_remove("./data/files/".$torrent["info"]["name"]);
    @unlink("./data/".$torrent["sha1"].".status");
    flock($fp,LOCK_UN);
    fclose($fp);    
  }

  function spawn($torrentfile) {
    $update_interval = 15; //Seconds between updates
    $fp = fopen("./data/spawnlock","wb");
    flock($fp,LOCK_EX);    
    $torrentfile = "./torrents/$torrentfile";
    $torrent = bdecode(file_get_contents($torrentfile));
    if (!is_array($torrent)) {
      flock($fp,LOCK_UN);
      fclose($fp);
      return false;
    }
    $stats = $this->getSpawnStatus($torrent["sha1"],$update_interval);
    
    if (count($stats)==0) {
      $statusfile = "data/".$torrent["sha1"].".status";
      $savein = "data/files/";
      $command = $this->python." data/bt/btdownloadbg.py \"$torrentfile\" --statusfile $statusfile --display_interval $update_interval --save_in $savein 2>&1";
      passthru($command);
      flock($fp,LOCK_UN);
      fclose($fp);
      return true;
    } elseif (!$stats["Running"]) {
      $this->stop_by_pid($stats["process id"]);

      $statusfile = "data/".$torrent["sha1"].".status";
      $savein = "data/files/";
      $command = $this->python." data/bt/btdownloadbg.py \"$torrentfile\" --statusfile $statusfile --display_interval $update_interval --save_in $savein 2>&1";
      passthru($command);
      flock($fp,LOCK_UN);
      fclose($fp);
      return true;
    } else {//It's already running
      flock($fp,LOCK_UN);
      fclose($fp);
      return false;
    }
  }   
}